// card.c

// 3-10: Now thread-safe(r) with help from Jon Watte (hplus@b500.com)

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <string.h>
#include <stdio.h>
#include "GemTekRadio.h"

__declspec(dllexport)
// Implemented to comply with Be driver interface; they don't do much
status_t init_hardware(void);	// Mutes card
status_t init_driver(void);
void uninit_driver(void);

//status_t r_open( const char *, uint32, void** );

const char ** publish_devices(void);
device_hooks * find_device(const char *);

///////////////////////////////////////////////
// Read from card
unsigned short read_card(int sub_port);
// Write to card
void write_card(int sub_port, unsigned short value);
// I/O Helper functions
void zero( void );
void one( void );

sem_id g_sem;

struct gemtek_device
{
	int port;
	unsigned long curfreq;
	int muted;
};

static struct gemtek_device gemtek_unit;

static void gemtek_mute(struct gemtek_device *dev)
{
    if(dev->muted)
		return;
	write_card(0, 0x10);
	dev->muted = 1;
}

static void gemtek_unmute(struct gemtek_device *dev)
{
	if(dev->muted == 0)
		return;
	write_card(0, 0x20);
	dev->muted = 0;
}

void zero(void)
{
	write_card(0, 0x04);
	spin(5);
	write_card(0, 0x05);
	spin(5);
}

void one(void)
{
	write_card(0, 0x06);
	spin(5);
	write_card(0, 0x07);
	spin(5);
}

static int gemtek_setfreq(struct gemtek_device *dev, unsigned long freq)
{
	int i;

/*        freq = 78.25*((float)freq/16000.0 + 10.52); */

	freq /= 16;
	freq += 10520;
	freq *= 7825;
	freq /= 100000;

	/* 2 start bits */
	write_card(0, 0x03);
	spin(5);
	write_card(0, 0x07);
	spin(5);

        /* 28 frequency bits (lsb first) */
	for (i = 0; i < 14; i++)
		if (freq & (1 << i))
			one();
		else
			zero();
        /* 36 unknown bits */
	for (i = 0; i < 11; i++)
		zero();
	one();
	for (i = 0; i < 4; i++)
		zero();
	one();
	zero();

	/* 2 end bits */
	write_card(0, 0x03);
	spin(5);
	write_card(0, 0x07);
	spin(5);

	return 0;
}

int gemtek_getsigstr(struct gemtek_device *dev)
{
	read_card(0);
	spin(5);
	if (read_card(0) & 8)		/* bit set = no signal present */
		return 0;
	return 1;		/* signal present */
}

/* init_hardware - called once the first time the driver is loaded */
status_t
init_hardware (void)
{
	return B_OK;
}

status_t
init_driver (void)
{
	g_sem = create_sem( 1, "GemTekRadio lock" );
	return (g_sem < 0) ? g_sem : B_OK;
}

void
uninit_driver (void)
{
	delete_sem( g_sem );
}

static status_t
r_open (const char *name, uint32 flags, void** cookie)
{
	return B_OK;
}

static status_t
r_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	*num_bytes = 0;				/* tell caller nothing was read */
	return B_IO_ERROR;
}


/* r_write: handle write() calls.
	3/10/99: Functionality temporarily removed because it didn't work right */
static status_t
r_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	return B_OK;
}


/* r_control: decides whether to tune or mute the card */
static status_t
r_control (void* cookie, uint32 op, void* arg, size_t len)
{
	gemtek_getsigstr(&gemtek_unit);

	switch(op) {
	case RADIO_TUNE:
	{
		unsigned long val = *(unsigned long*)arg;
		gemtek_mute(&gemtek_unit);
		gemtek_setfreq(&gemtek_unit,val);
		gemtek_setfreq(&gemtek_unit,val);
		gemtek_unmute(&gemtek_unit);
	}
	break;
	case RADIO_ON:
		gemtek_unmute(&gemtek_unit);
		break;

	case RADIO_OFF:
		gemtek_mute(&gemtek_unit);
		break;
	default:
		return EINVAL;
	}
	return B_OK;
}

static status_t
r_close (void* cookie)
{
	return B_OK;
}

static status_t
r_free (void* cookie)
{
	return B_OK;
}

/* null-terminated array of device names */
static const char *my_device_name[] = {
	"misc/GemTekRadio",
	NULL
};

device_hooks my_device_hooks = {
	r_open, 			/* -> open entry point */
	r_close, 			/* -> close entry point */
	r_free,				/* -> free cookie */
	r_control, 			/* -> control entry point */
	r_read,				/* -> read entry point */
	r_write				/* -> write entry point */
};

/* publish_devices - return a null-terminated array of devices
	supported by this driver */
const char**
publish_devices()
{
	return my_device_name;
}

/* find_device - return ptr to device hooks structure */
device_hooks*
find_device(const char* name)
{
	return &my_device_hooks;
}

/* reads a byte in from the card */
unsigned short read_card(int sub_port)
{
   return read_io_8(RADIO_BASE+sub_port);
}

/* writes a byte to the card */
void write_card(int sub_port, unsigned short value)
{
   write_io_8(RADIO_BASE+sub_port,value);
}
